package djmil.cordacheckers.cordaclient;

import static java.util.Objects.requireNonNull;

import java.util.List;
import java.util.UUID;
import java.util.concurrent.TimeUnit;

import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import djmil.cordacheckers.cordaclient.dao.HoldingIdentity;
import djmil.cordacheckers.cordaclient.dao.Piece;
import djmil.cordacheckers.cordaclient.dao.GameBoard;
import djmil.cordacheckers.cordaclient.dao.GameBoardCommand;
import djmil.cordacheckers.cordaclient.dao.GameProposal;
import djmil.cordacheckers.cordaclient.dao.VirtualNode;
import djmil.cordacheckers.cordaclient.dao.VirtualNodeList;
import djmil.cordacheckers.cordaclient.dao.flow.RequestBody;
import djmil.cordacheckers.cordaclient.dao.flow.ResponseBody;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.GameProposalCreateReq;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.GameProposalCreateRes;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.GameProposalListRes;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.Empty;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.GameBoardCommandReq;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.GameBoardCommandRes;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.GameBoardListRes;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.GameProposalCommandAcceptRes;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.GameProposalCommandReq;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.GameProposalCommandRes;

@Service
public class CordaClient {
    private final RestTemplate restTemplate;
    private final ObjectMapper jsonMapper;

    public CordaClient(RestTemplateBuilder restTemplateBuilder , ObjectMapper jsonMapper) {
        this.restTemplate = restTemplateBuilder.build();
        this.jsonMapper = jsonMapper;
    }

    public List<VirtualNode> getVirtualNodeList() {
        ResponseEntity<VirtualNodeList> resp = this.restTemplate.exchange(
            "/virtualnode",
            HttpMethod.GET, 
            null, 
            VirtualNodeList.class
        );

        if (resp.getStatusCode() != HttpStatus.OK) {
            throw new RuntimeException("CordaClient.getVirtualNodeList: unexpected request status "+resp.getStatusCode()) ;
        }

        return requireNonNull(
                resp.getBody(),
                "CordaClient.getVirtualNodeList: empty getBody()"
            ).virtualNodes();
    }

    /**
     * Obtain list of unconsumed (active) GameProposals 
     * @param myHoldingIdentity
     * @return GameProposals list in JSON form
     */
    public List<GameProposal> gameProposalList(HoldingIdentity myHoldingIdentity) {

        final RequestBody requestBody = new RequestBody(
            "gp.list-" + UUID.randomUUID(),
            "djmil.cordacheckers.gameproposal.ListFlow",
            new Empty()
            );

        final GameProposalListRes listFlowResult = cordaFlowExecute(
            myHoldingIdentity,
            requestBody,
            GameProposalListRes.class
            );

        if (listFlowResult.failureStatus() != null) {
            System.out.println("GameProposalCreateFlow failed: " + listFlowResult.failureStatus());
            throw new RuntimeException("GameProsal: CreateFlow execution has failed");
        }

        return listFlowResult.successStatus();
    }

    public UUID gameProposalCreate(
        HoldingIdentity issuer,
        HoldingIdentity acquier,
        Piece.Color acquierColor,
        String message
    ) throws JsonMappingException, JsonProcessingException {
        final RequestBody requestBody = new RequestBody(
            "gp.create-" + UUID.randomUUID(),
            "djmil.cordacheckers.gameproposal.CreateFlow",
            new GameProposalCreateReq(
                acquier.x500Name(),
                acquierColor,
                message
                )
            );

        final GameProposalCreateRes createResult = cordaFlowExecute(
            issuer,
            requestBody,
            GameProposalCreateRes.class
            );

        if (createResult.failureStatus() != null) {
            System.out.println("GameProposal.CreateFlow failed: " + createResult.failureStatus());
            throw new RuntimeException("GameProsal: CreateFlow execution has failed");
        }

        return createResult.successStatus();
    }

    public String gameProposalReject(
        HoldingIdentity myHoldingIdentity,
        UUID gameProposalUuid
    ) {
        final RequestBody requestBody = new RequestBody(
            "gp.reject-" +UUID.randomUUID(),
            "djmil.cordacheckers.gameproposal.CommandFlow",
            new GameProposalCommandReq(
                gameProposalUuid.toString(),
                GameProposalCommandReq.Command.REJECT
                )
            );

        final GameProposalCommandRes actionResult = cordaFlowExecute(
            myHoldingIdentity,
            requestBody,
            GameProposalCommandRes.class
            );

        if (actionResult.failureStatus() != null) {
            System.out.println("GameProposal.CommandFlow failed: " + actionResult.failureStatus());
            throw new RuntimeException("GameProsal: CommandFlow execution has failed");
        }

        return actionResult.successStatus();
    }

    public GameBoard gameProposalAccept(
        HoldingIdentity myHoldingIdentity,
        UUID gameProposalUuid
    ) {
        final RequestBody requestBody = new RequestBody(
            "gp.accept-" +UUID.randomUUID(),
            "djmil.cordacheckers.gameproposal.CommandFlow",
            new GameProposalCommandReq(
                gameProposalUuid.toString(),
                GameProposalCommandReq.Command.ACCEPT
                )
            );

        final GameProposalCommandAcceptRes actionResult = cordaFlowExecute(
            myHoldingIdentity,
            requestBody,
            GameProposalCommandAcceptRes.class
            );

        if (actionResult.failureStatus() != null) {
            System.out.println("GameProposal.CommandFlow failed: " + actionResult.failureStatus());
            throw new RuntimeException("GameProsal: CommandFlow execution has failed");
        }

        return actionResult.successStatus();
    }

    public List<GameBoard> gameBoardList(HoldingIdentity holdingIdentity) {

        final RequestBody requestBody = new RequestBody(
            "gb.list-" + UUID.randomUUID(),
            "djmil.cordacheckers.gameboard.ListFlow",
            new Empty()
            );

        final GameBoardListRes listFlowResult = cordaFlowExecute(
            holdingIdentity,
            requestBody,
            GameBoardListRes.class
            );

        if (listFlowResult.failureStatus() != null) {
            System.out.println("GameBoard.ListFlow failed: " + listFlowResult.failureStatus());
            throw new RuntimeException("GameBoard: ListFlow execution has failed");
        }

        return listFlowResult.successStatus();
    }

    public GameBoard gameBoardCommand(
        HoldingIdentity myHoldingIdentity,
        UUID gameBoardUuid,
        GameBoardCommand command
    ) {
        final RequestBody requestBody = new RequestBody(
            "gb.command-" +command.getType() +UUID.randomUUID(),
            "djmil.cordacheckers.gameboard.CommandFlow",
            new GameBoardCommandReq(
                gameBoardUuid,
                command
                )
            );

        final GameBoardCommandRes moveResult = cordaFlowExecute(
            myHoldingIdentity,
            requestBody,
            GameBoardCommandRes.class 
            );

        if (moveResult.failureStatus() != null) {
            System.out.println("GameBoard.MoveFlow failed: " + moveResult.failureStatus());
            throw new RuntimeException("GameBoard: MoveFlow execution has failed");
        }

        return moveResult.successStatus();
    }

    private <T> T cordaFlowExecute(HoldingIdentity holdingIdentity, RequestBody requestBody, Class<T> flowResultType) {

        try {
            final String requestBodyJson = this.jsonMapper.writeValueAsString(requestBody);

            final ResponseBody startedFlow = cordaFlowPost(
                holdingIdentity,
                requestBodyJson
            );

            final String flowResult = cordaFlowPoll(startedFlow);

            return this.jsonMapper.readValue(flowResult, flowResultType);
        } 
        catch (Exception e) {
            throw new RuntimeException("Unable to perform "+requestBody.flowClassName()
                                        +". Reason "+e.getMessage());
        }
    }

    private ResponseBody cordaFlowPost(HoldingIdentity holdingIdentity, String requestBodyJson) {
		final HttpEntity<String> request = new HttpEntity<>(requestBodyJson);

        final ResponseEntity<ResponseBody> responce = this.restTemplate.exchange(
            "/flow/" + holdingIdentity.shortHash(),
            HttpMethod.POST,
            request, 
            ResponseBody.class
        );
        
        if (!responce.getStatusCode().is2xxSuccessful()) {
            throw new RuntimeException("CordaClient.startCordaFlow: unexpected request status "
                                        +responce.getStatusCode()) ;
        }

        final ResponseBody responseBody = requireNonNull(
            responce.getBody(),
            "CordaClient.startCordaFlow: empty getBody()"
        );

        if (!responseBody.isFlowStarted()) {
            throw new RuntimeException("CordaClient.startCordaFlow: Unexpected status: "
                +responseBody.flowStatus() + ": " 
                +responseBody.flowError());
        }

        return responseBody;
    }

    private String cordaFlowPoll(ResponseBody startedFlow) throws InterruptedException {

        for (int retry = 0; retry < 6; retry++) {
            // Give Corda cluster some time to process our request
            TimeUnit.SECONDS.sleep(retry*retry +2); // 2 3 6 10 18 27 sec

            final ResponseEntity<ResponseBody> responce = this.restTemplate.exchange(
                "/flow/" 
                    + startedFlow.holdingIdentityShortHash()+"/"
                    + startedFlow.clientRequestId(),
                HttpMethod.GET,
                null, 
                ResponseBody.class
            );
          
            if (responce.getStatusCode() != HttpStatus.OK) {
                throw new RuntimeException("CordaClient.cordaFlowPoll: unexpected request status "
                                            +responce.getStatusCode()) ;
            }

            final ResponseBody responseBody = requireNonNull(
                responce.getBody(),
                "CordaClient.cordaFlowPoll: empty getBody()"
            );

            if (responseBody.flowError() != null) {
                throw new RuntimeException("CordaClient.cordaFlowPoll: flow execution error: " 
                                            +responseBody.flowError());
            }

            if (responseBody.isFlowCompleted() && responseBody.flowResult() != null) {
                return responseBody.flowResult();
            }
        }

        throw new RuntimeException ("CordaClient.cordaFlowPoll: retry limit");
    }

}
