package djmil.cordacheckers.cordaclient;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

import java.util.List;
import java.util.UUID;

import javax.naming.InvalidNameException;

import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonMappingException;

import djmil.cordacheckers.cordaclient.dao.CordaState;
import djmil.cordacheckers.cordaclient.dao.GameBoard;
import djmil.cordacheckers.cordaclient.dao.GameBoardCommand;
import djmil.cordacheckers.cordaclient.dao.GameProposal;
import djmil.cordacheckers.cordaclient.dao.Piece;
import djmil.cordacheckers.cordaclient.dao.VirtualNode;
import djmil.cordacheckers.user.HoldingIdentityResolver;

@SpringBootTest
public class CordaClientTest {
    @Autowired
    CordaClient cordaClient;

    @Autowired
    HoldingIdentityResolver holdingIdentityResolver;

    @Test
    void testGetVirtualNodeList() throws InvalidNameException {
        List<VirtualNode> vNodes = cordaClient.getVirtualNodeList();

		assertThat(vNodes.size()).isEqualTo(5); // default vNode config for CSDE
    }

    @Test
    void testGameProposalList() throws JsonProcessingException {
        List<GameProposal> gpList = cordaClient.gameProposalList(
            holdingIdentityResolver.getByUsername("Alice"));

        System.out.println("testListGameProposals\n"+ gpList);
    }

    @Test
    void testGameProposalCreate() throws JsonMappingException, JsonProcessingException {
        final String gpIssuer = "alice";
        final String gpAcquier = "bob";
        final Piece.Color gpAcquierColor = Piece.Color.WHITE;
        final String gpMessage = "GameProposal create test";

        final UUID createdGpUuid = cordaClient.gameProposalCreate(
            holdingIdentityResolver.getByUsername(gpIssuer),
            holdingIdentityResolver.getByUsername(gpAcquier),
            gpAcquierColor,
            gpMessage);

        List<GameProposal> gpListSender = cordaClient.gameProposalList(
            holdingIdentityResolver.getByUsername(gpIssuer));

        assertThat(findByUuid(gpListSender, createdGpUuid)).isNotNull();

        List<GameProposal> gpListReceiver = cordaClient.gameProposalList(
            holdingIdentityResolver.getByUsername(gpAcquier));

        GameProposal gp;
        assertThat(gp = findByUuid(gpListReceiver, createdGpUuid)).isNotNull();

        assertThat(gp.acquier()).isEqualToIgnoringCase(gpAcquier);
        assertThat(gp.issuer()).isEqualToIgnoringCase(gpIssuer);
        assertThat(gp.acquierColor()).isEqualByComparingTo(gpAcquierColor);
        assertThat(gp.message()).isEqualTo(gpMessage);
    }

    @Test
    void testGameProposalReject() throws JsonMappingException, JsonProcessingException {
        final String gpIssuer = "alice";
        final String gpAcquier = "bob";
        final Piece.Color gpReceiverColor = Piece.Color.WHITE;
        final String gpMessage = "GameProposal REJECT test";

        final UUID gpUuid = cordaClient.gameProposalCreate(
            holdingIdentityResolver.getByUsername(gpIssuer),
            holdingIdentityResolver.getByUsername(gpAcquier),
            gpReceiverColor,
            gpMessage);

        System.out.println("Create GP UUID "+ gpUuid);

        assertThatThrownBy(() -> {  // Issuer can not reject
            cordaClient.gameProposalReject(
                holdingIdentityResolver.getByUsername(gpIssuer),
                gpUuid);
        });

        final String rejectRes = cordaClient.gameProposalReject(
            holdingIdentityResolver.getByUsername(gpAcquier),
            gpUuid);

        assertThat(rejectRes).isEqualToIgnoringCase("REJECTED");

        List<GameProposal> gpListSender = cordaClient.gameProposalList(
            holdingIdentityResolver.getByUsername(gpIssuer));

        assertThat(findByUuid(gpListSender, gpUuid)).isNull();

        List<GameProposal> gpListReceiver = cordaClient.gameProposalList(
            holdingIdentityResolver.getByUsername(gpAcquier));

        assertThat(findByUuid(gpListReceiver, gpUuid)).isNull();

        // GameProposal can not be rejected twice 
        assertThatThrownBy(() -> {
            cordaClient.gameProposalReject(
                holdingIdentityResolver.getByUsername(gpIssuer),
                gpUuid);
        });
    }

    @Test
    void testGameProposalAccept() throws JsonMappingException, JsonProcessingException {
        final String gpIssuer = "alice";
        final String gpAcquier = "bob";
        final Piece.Color gpAcquierColor = Piece.Color.WHITE;
        final String gpMessage = "GameProposal ACCEPT test";

        final UUID gpUuid = cordaClient.gameProposalCreate(
            holdingIdentityResolver.getByUsername(gpIssuer),
            holdingIdentityResolver.getByUsername(gpAcquier),
            gpAcquierColor,
            gpMessage
        );

        System.out.println("New GameProposal UUID "+ gpUuid);

        assertThatThrownBy(() -> { // Issuer can not accept
            cordaClient.gameProposalAccept(
                holdingIdentityResolver.getByUsername(gpIssuer),
                gpUuid);
        });

        final GameBoard gbState = cordaClient.gameProposalAccept(
            holdingIdentityResolver.getByUsername(gpAcquier),
            gpUuid);

        System.out.println("New GameBoard UUID "+gbState);

        List<GameBoard> gbListIssuer = cordaClient.gameBoardList(
            holdingIdentityResolver.getByUsername(gpIssuer));

        GameBoard gbAlice = findByUuid(gbListIssuer, gbState.id());
        assertThat(gbAlice).isNotNull();
        assertThat(gbAlice.opponentName()).isEqualToIgnoringCase(gpAcquier);
        assertThat(gbAlice.opponentColor()).isEqualByComparingTo(gpAcquierColor);
        assertThat(gbAlice.opponentMove()).isEqualTo(true);


        List<GameBoard> gbListAcquier = cordaClient.gameBoardList(
        holdingIdentityResolver.getByUsername(gpAcquier));

        GameBoard bgBob = findByUuid(gbListAcquier, gbState.id());
        assertThat(bgBob).isNotNull();
        assertThat(bgBob.opponentName()).isEqualToIgnoringCase(gpIssuer);
        assertThat(bgBob.opponentColor()).isEqualByComparingTo(Piece.Color.BLACK);
        assertThat(bgBob.opponentMove()).isEqualTo(false);

        assertThatThrownBy(() -> {   // GameProposal can not be accepted twice 
            cordaClient.gameProposalAccept(
                holdingIdentityResolver.getByUsername(gpAcquier),
                gpUuid);
        });
    }

    @Test
    void testGameBoardList() throws JsonMappingException, JsonProcessingException {
        List <GameBoard> gbList = cordaClient.gameBoardList(
            holdingIdentityResolver.getByUsername("bob"));

        System.out.println("GB list: " +gbList);
    }

    @Test
    void testGameBoardSurrender() throws JsonMappingException, JsonProcessingException {
        final var hiAlice = holdingIdentityResolver.getByUsername("alice");
        final var hiBob   = holdingIdentityResolver.getByUsername("bob");
        final var bobColor = Piece.Color.WHITE;

        final UUID gpUuid = cordaClient.gameProposalCreate(
            hiAlice, hiBob,
            bobColor, "GameBoard SURRENDER test"
        );

        System.out.println("New GameProposal UUID "+ gpUuid);

        final GameBoard gbState = cordaClient.gameProposalAccept(
            hiBob, gpUuid
        );

        System.out.println("New GameBoard UUID "+ gbState.id());

        final GameBoard gbSurrender = cordaClient.gameBoardCommand(
            hiBob, gbState.id(), 
            new GameBoardCommand(GameBoardCommand.Type.SURRENDER)
            );

        System.out.println("SURRENDER GB: "+gbSurrender);
    }

    private <T extends CordaState> T findByUuid(List<T> statesList, UUID uuid) {
        for (T state : statesList) {
            if (state.id().compareTo(uuid) == 0)
                return state;
        };
        return null;
    }
}
