package djmil.cordacheckers.gameproposal;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.FlowResult;
import djmil.cordacheckers.contracts.GameProposalCommand;
import djmil.cordacheckers.states.GameProposalState;
import djmil.cordacheckers.states.Piece;
import net.corda.v5.application.flows.ClientRequestBody;
import net.corda.v5.application.flows.ClientStartableFlow;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.FlowEngine;
import net.corda.v5.application.marshalling.JsonMarshallingService;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.crypto.SecureHash;
import net.corda.v5.ledger.common.NotaryLookup;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;
import net.corda.v5.membership.MemberInfo;
import net.corda.v5.membership.NotaryInfo;

import static java.util.Objects.requireNonNull;

import java.time.Duration;
import java.time.Instant;
import java.util.Arrays;
import java.util.UUID;

public class CreateFlow implements ClientStartableFlow{

     private final static Logger log = LoggerFactory.getLogger(CreateFlow.class);

     @CordaInject
     public JsonMarshallingService jsonMarshallingService;

     @CordaInject
     public MemberLookup memberLookup;

     @CordaInject
     public NotaryLookup notaryLookup;

     @CordaInject
     public UtxoLedgerService ledgerService;

     @CordaInject
     public FlowEngine flowEngine;

     @Suspendable
     @Override
     public String call(ClientRequestBody requestBody) {
          try {
               log.info("flow: Create Game Proposal");
               final GameProposalCommand command = GameProposalCommand.CREATE;

               final GameProposalState newGameProposal = buildGameProposalStateFrom(requestBody);
               
               final UtxoSignedTransaction trx = prepareSignedTransaction(newGameProposal);

               final SecureHash trxId = this.flowEngine
                    .subFlow(new CommitSubFlow(trx, command.getRespondent(newGameProposal)));

               return new FlowResult(newGameProposal.getId(), trxId)
                    .toJsonEncodedString(jsonMarshallingService);
          } 
          catch (Exception e) {
               log.warn("CreateGameProposal flow failed to process utxo request body " + requestBody + 
                         " because: " + e.getMessage());
               return new FlowResult(e).toJsonEncodedString(jsonMarshallingService);
          }
     }

     @Suspendable
     private GameProposalState buildGameProposalStateFrom(ClientRequestBody requestBody) {
          CreateFlowArgs args = requestBody.getRequestBodyAs(jsonMarshallingService, CreateFlowArgs.class);

          Piece.Color opponentColor = Piece.Color.valueOf(args.opponentColor);
          if (opponentColor == null) {
               throw new RuntimeException("Allowed values for opponentColor are: "
                    + Piece.Color.WHITE.name() +", "  + Piece.Color.BLACK.name() 
                    + ". Actual value was: " + args.opponentColor);
          }

          MemberInfo myInfo = memberLookup.myInfo();
          MemberInfo opponentInfo = requireNonNull(
               memberLookup.lookup(MemberX500Name.parse(args.opponentName)),
               "MemberLookup can't find opponentName specified in flow arguments: " + args.opponentName
          );

          return new GameProposalState(
               myInfo.getName(),
               opponentInfo.getName(),
               opponentColor,
               args.message,
               UUID.randomUUID(),
               Arrays.asList(myInfo.getLedgerKeys().get(0), opponentInfo.getLedgerKeys().get(0))
          );
     }

     @Suspendable
     private UtxoSignedTransaction prepareSignedTransaction(GameProposalState outputGameProposalState) {     
          final NotaryInfo notary = notaryLookup.getNotaryServices().iterator().next();

          return ledgerService.createTransactionBuilder()
               .setNotary(notary.getName())
               .setTimeWindowBetween(Instant.now(), Instant.now().plusMillis(Duration.ofDays(1).toMillis()))
               .addOutputState(outputGameProposalState)
               .addCommand(GameProposalCommand.CREATE)
               .addSignatories(outputGameProposalState.getParticipants())
               .toSignedTransaction();
     }
}
